#Searching the access nodes for transit node routing
# Programmed by Mikio Kubo 
# Copyright supply chain optimization labo. Shortest Path Project (2008)

from pylab import *
from networkx import *  # import networkx after pylab 
import numpy 
      
def ReadDimacs(filename):
    '''Read the data from DIMACS shortest path format
       that can be downloaded from DIMACS site'''
    
    f=open(filename+'.co', 'r')
    f2=open(filename+'.gr', 'r')

    # read data from file
    line=f.readline()
    x_cord=[]
    y_cord=[]
    G=XGraph()
    G.position={}
        
    while line.find("v 1")==-1:
        line=f.readline()

    (dum,i,x,y)=line.split()
    x_cord.append(float(x))
    y_cord.append(float(y))

    while 1:
        line=f.readline()
        if line.find("v")==0:
            (dum,i,x,y)=line.split()
            x_cord.append(float(x))
            y_cord.append(float(y))
        else:
            break

    #edge info
    edge_info=[]
    line=f2.readline()

    for i in range(6):
        line=f2.readline()
        #print(line)
        
    while 1:
        line=f2.readline()
        if line.find("a")==0:
            (dum,i,j,length)=line.split()
            edge_info.append( (int(i)-1,int(j)-1,int(length)) )
            #print ("edge info",i,j,length)
        else:
            break

    #print(x_cord)
    #print(edge_info)
   
    
    minx=float(min(x_cord))
    maxx=float(max(x_cord))
    xwidth= maxx-minx+1

    miny=float(min(y_cord))
    maxy=float(max(y_cord))
    ywidth=maxy-miny+1

    #print minx,maxx,miny,maxy

    # add nodes after scaling coordinates
    n=len(x_cord)
    for i in range(n):
        G.add_node(i)  
        x=(x_cord[i]-minx)/xwidth
        y=(y_cord[i]-miny)/ywidth
        G.position[i]=(x,y)   
        #print i,G.position[i]

    print "number of nodes=",n

    # add edges
    m=len(edge_info)
    C=0
    for (i,j,length)in edge_info:
        G.add_edge(i,j,length)
        #print "add edge",i,j,length
        if C<length: C=length
    C+=1

    print "number of edges=",m

    f.close()
    f2.close()
    
    return G,n,m,C

def DialSubForAccess(n,G,settled,reached,covered,transit,dist,prev,C,source,sink,Access,TransitGraph):
    ''' Dial method subroutine
     for finding access nodes transit'''
    circular=[[-1] for i in range(C) ] #-1 represents the list is empty
    circular[0].append(source)
    reached[source]=1
    prev[source]=-1
    current=0
    num_uncovered=1
    covered[source]=0

    scanned=0
    Searched=set([])
    while 1:
      
        first=current
        while 1:
            v=circular[current][-1] #v is the node with minimum potential
            if v>=0:
                break
            #the current list is empty, so we increment it
            current+=1
            current= current%C
            if current==first:
                #the circular list is empty, so we quit
                v=-1
                break
            
        if v==-1: #the circular list is empty, so we quit
            break
        circular[current].pop()
        settled[v]=1
        #if v==sink:
        #    break

        # transit node (except source node) is covered when settled
        if transit[v]==1:
            if v !=source and covered[v]==0: 
                covered[v]=1
                num_uncovered-=1
        else:
            if covered[v]==0: 
                num_uncovered -=1

        if num_uncovered==1 and v !=source:
            break #all the nodes in the heap are covered 

        scanned+=1
        #if scanned>=3000:
        #    break
        
        #print "scanning node",v,num_uncovered
        #Uncovered=[]
        #for i in Searched:
        #    if covered[i]==0 and settled[i]==0:
        #        Uncovered.append(i)
        #print "Uncovered Nodes=",len(Uncovered)
    
        for w in G.neighbors(v): #scan operation
            if settled[w]==0:
                Searched.add(w)  
                temp=dist[v]+G.get_edge(v,w)
                if reached[w]== 0: #w is not in the circular list
                    reached[w]=1
                    dist[w]=temp
                    prev[w]=v
                    circular[temp%C].append(w)

                    if covered[v]==1:
                        covered[w]=1
                    else:
                        num_uncovered+=1
                        
                else: #w is in the circular list
                    if temp<dist[w]:
                        circular[dist[w]%C].remove(w)
                        circular[temp%C].append(w)
                        dist[w]=temp
                        prev[w]=v

                        if covered[v]==1:
                            if covered[w]==0:
                                covered[w]=1
                            #print "node ",w," is covered by node",v
                                num_uncovered-=1
                        else:
                            if covered[w]==1:
                                covered[w]=0
                                num_uncovered+=1

    Covered=[] #covering nodes
    for v in Searched:
        if transit[v]==1 and covered[prev[v]]==0:
            Covered.append(v)
            TransitGraph.add_edge(source,v,dist[v])
            
    Settled=[]  #settled but not covered
    for v in Searched:
        if settled[v]==1 and covered[v]==0:
            Settled.append(v)
            Access[v].append( (source,dist[v]))

    #clear arrays used in the sub
    for v in Searched:
        settled[v]=0
        reached[v]=0
        covered[v]=0
        dist[v]=0
        prev[v]=0
        
    #print Access    
    #draw(G,G.position,nodelist=Searched,node_size=20,with_labels=None,width=1,node_color='y',edge_color='b')
    #draw(G,G.position,nodelist=Covered,node_size=50,with_labels=None,width=1,node_color='y',edge_color='b')
    #draw(G,G.position,nodelist=Uncovered,node_size=70,with_labels=None,width=1,node_color='g',edge_color='b')
    #draw(G,G.position,nodelist=Settled,node_size=10,with_labels=None,width=1,node_color='r',edge_color='b')

    print "Searched=",len(Searched)
                        
if __name__ == "__main__":
    filename="DC.tmp"
    #filename="USA-road-d.NY"
    #filename="NJ.tmp"
    G,n,m,C = ReadDimacs(filename)
    print "number of nodes in the largest connected componet=",n

    transit= numpy.zeros(n,numpy.int0) 
    f=open(filename+'.transit', 'r')
    TransitAll=[]
    for node in G.nodes_iter():
        line=f.readline()
        temp=line.split()
        transit[node]=int(temp[-1])
        if transit[node]==1:
            #print node
            TransitAll.append(node)

    draw(G,G.position,nodelist=TransitAll,node_size=100,with_labels=None,width=1,node_color='b',edge_color='g')

    settled= numpy.zeros(n,numpy.int0) #=1 if node is settled (or has an eternal label or is scanned)
    reached= numpy.zeros(n,numpy.int0) #=1 if node is heap or list (or has a finite potential)    
    covered =numpy.zeros(n,numpy.int0) #=1 if node is covered by transit node (initialized to 0)

    dist= numpy.zeros(n,numpy.int0)
    prev= numpy.zeros(n,numpy.int0)

    Access=[[] for i in range(n)] #list of (access node, distance)
    TransitGraph= XGraph()

    #source=TransitAll[0]
    #print "finding access point to",source
    sink =-1 #if sink =-1 then find all shortest paths from source to other nodes
    #SourceList=[source]
    #draw(G,G.position,nodelist=SourceList,node_size=100,with_labels=None,width=1,node_color='r',edge_color='g')

    count=1
    for source in TransitAll:
        print "finding access point to",source,count
        count+=1
        #if count>=10:
        #    break
        DialSubForAccess(n,G,settled,reached,covered,transit,dist,prev,C,source,sink,Access,TransitGraph)

##    if sink>=0:
##        path=[sink]
##        w=sink
##        while 1:
##            v=prev[w]
##            path.append(v)
##            if v==source:
##                break
##            w=v
##        path.reverse()
##        print "Path Length=",dist[sink]
##        print "Optimal Path",path
##
##        G.clear()
##        G.add_path(path)
##        draw(G,G.position,node_size=1000/n,with_labels=None,width=3,node_color='b',edge_color='r')
##
##    else:
##        print "Distance Array=",dist
##        print "Prev Array",prev

    #output the transit graph
    write_multiline_adjlist(TransitGraph, filename+'.tg')
    #stop
    #output the information of access nodes
    f=open(filename+'.access', 'w')
    for node in G.nodes_iter():
        f.write(str(node)+"\n")
        for item in Access[node]:
             f.write(str(item)+"\t")
        f.write("\n")
    f.close()

    #draw(G,G.position,node_size=1000/n,with_labels=None,width=1,node_color='b',edge_color='g')
    #draw(TransitGraph,G.position,node_size=1000/n,with_labels=None,width=1,node_color='b',edge_color='r')
    #show()    

    
                
                    
                

        

        
